function [PHI,E,Am,L] = cotan_LB(M,n_eigenvalues)
% COTAN_LB Calculates the mass and stiffness matrix and n_eigenvalues
% many eigenvectors and values of the the composition Am^-1 * L using
% the generalized eigenvalue solver of Matlab. 
% vertices - n x 3
% faces - m x 3
% PHI - n x n_eigenvalues, eigenvectors
% E - eigenvalues
% Am - n x n, lumped mass matrix
% L - n x n, stiffness matrix
%
% written by Emanuele Rodola

vertices = M.VERT;
faces = M.TRIV;

num_vertices = size(vertices,1);
num_faces = size(faces,1);

%% detect boundary vertices

% Calculate the (directed) adjacency matrix. adjacency_matrix(m,n) = 1 if the oriented
% boundary of a triangle contains the directed edge from vertex m to vertex
% n, and 0 otherwise. This matrix is not quite symmetric due to boundary edges.
adjacency_matrix = sparse([faces(:,1); faces(:,2); faces(:,3)], ...
                         [faces(:,2); faces(:,3); faces(:,1)], ...
    	                 ones(3 * num_faces, 1), ...
                         num_vertices, num_vertices, 3 * num_faces);
if any(any(adjacency_matrix > 1))
    error('Triangles must be oriented consistently.')
end
clear adjacency_matrix;


%% compute cotan LB matrix

fprintf('Computing Laplace-Beltrami operator...');

% first compute inner face angles and squared edge lengthes
pp = zeros(num_faces,3);
qq = zeros(num_faces,3);
angles = 0*faces;
squared_edge_length = 0*faces;

for i=1:3
    i1 = mod(i-1,3)+1;
    i2 = mod(i  ,3)+1;
    i3 = mod(i+1,3)+1;
    pp = vertices(faces(:,i2),:) - vertices(faces(:,i1),:);
    qq = vertices(faces(:,i3),:) - vertices(faces(:,i1),:);
    % normalize the vectors
    pp = pp ./ repmat( max(sqrt(sum(pp.^2,2)),eps), [1 3] );
    qq = qq ./ repmat( max(sqrt(sum(qq.^2,2)),eps), [1 3] );
    % compute angles
    angles(:,i1) = acos(sum(pp.*qq,2));
    %squared_edge_length(:,i1) = sum((vertices(faces(:,i2)) - vertices(faces(:,i3))).^2,2);
    squared_edge_length(:,i1) = sum((vertices(faces(:,i2),:) - vertices(faces(:,i3),:)).^2,2);
end
clear pp qq;

%then compute L
L = sparse(num_vertices,num_vertices);
for i=1:3
    i1 = mod(i-1,3)+1;
    i2 = mod(i  ,3)+1;
    i3 = mod(i+1,3)+1;
    L = L + sparse(faces(:,i1),faces(:,i2),-cot(angles(:,i3)),...
        num_vertices,num_vertices,num_faces);       
end

L = 1/2 * (L + L');
L = sparse(1:num_vertices,1:num_vertices,-sum(L,2),num_vertices,num_vertices,...
    num_vertices) + L;

fprintf('done. \n');
%% compute Voronoi areas for vertices (following the algorithm decribed in 
%   Discrete Differential-Geometry Operators for Triangulated 2-Manifolds, 
%   Mark Meyer, Mathieu Desbrun, Peter Schroeder and Alan H. Barr, VisMath 2002. 

fprintf('Computing area of vertex Voronoi cells...');

A = vertex_voronoi(M);
Am = sparse(1:num_vertices, 1:num_vertices,A);

clear angles faces_area;
fprintf('done. \n');

%% compute LB eigenvalues and eigenfunctions 

fprintf('Computing Laplace-Beltrami eigenfunctions... ');

% solve generalized eigenvalue problem
options.disp = 0;
[PHI,E] = eigs(L,Am,n_eigenvalues,-1e-5,options);
E = diag(E);
E=abs(real(E));
[E,idx] = sort(E);
PHI = PHI(:,idx);
PHI=real(PHI);

fprintf('done. \n');

end
